/*
** Module   : TrustedClientPlugin.java
** Abstract : Custom authentication plugin that allows a user to be
**            automatically logged in (without password!) by specifying the
**            target user id when starting P2J client. 
**
** Copyright (c) 2010-2013, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- -------------------------------- Description --------------------------------
** 001 SVL 20100916 Created the initial version. Allows a user to be
**                  automatically logged in.
** 002 SVL 20100923 Changed to conform the new Authenticator API.
** 003 OM  20131018 Updated to conform the new Authenticator API.
*/
package com.goldencode.p2j.security;

import com.goldencode.p2j.net.SessionListener;
import java.io.*;
import java.util.*;

/**
 * Custom authentication plugin that allows a user to be automatically logged
 * in (without password!) by specifiying the target user id when starting P2J
 * client. An example: <code>./client.sh -u userid</code>
 * <p>
 * In order to be able to log in without password, this plugin should be
 * specified as the auth plugin for the target users.
 * <p>
 * This plugin is **THE FORM OF DISABLING P2J SERVER'S SECURITY** for the
 * given set of users.
  *
 */
public class TrustedClientPlugin
implements Authenticator
{
   /** The value of "option" node from AuthPlugin for this Authenticator */
   private String authOption = null;

   /**
    * Implements client side custom authentication logic.
    *
    * @param    parameters
    *           Additional configuration parameters. This plugin uses
    *           "subjectId" parameter which corresponds the name of the user
    *           account to be used.
    * @param    code
    *           Not used.
    *
    * @return   Array of bytes to be transmitted to the server as the
    *           authorization input. Contains target user name.
    */
   @Override
   public byte[] clientAuthHook(Map<String, Object> parameters, int code)
   {
      String subjectId = "";
      if (parameters != null)
         subjectId = (String) parameters.get("subjectId");
      if (subjectId == null)
         subjectId = "";

      ByteArrayOutputStream bos = new ByteArrayOutputStream();
      DataOutputStream      dos = new DataOutputStream(bos);

      try
      {
         dos.writeUTF(subjectId);
      }
      catch (IOException ioe)
      {
         return new byte[0];
      }

      return bos.toByteArray();
   }

   /**
    * Finalizes any resources allocated during authentication by the client.
    */
   @Override
   public void clientFinalize()
   {
   }

   /**
    * Implements server side custom authentication logic.
    * <p>
    * Accepts the byte array produced by the client side authentication hook
    * as authentication input and custom parameters.
    *
    * @param    auth
    *           The authorization input from the client.
    * @param    entity
    *           The entity to be authenticated. Not used.
    *
    * @return   The result of the authentication processing.
    */
   @Override
   public AuthenticationResponse serverAuthHook(byte[] auth, String entity)
   {
      // parse the byte array input
      ByteArrayInputStream bis = new ByteArrayInputStream(auth);
      DataInputStream dis = new DataInputStream(bis);

      String userId;

      try
      {
         userId = dis.readUTF();
      }
      catch (IOException ioe)
      {
         return new AuthenticationResponse(null, AUTH_RESULT_UNSPECIFIED_FAILURE);
      }

      // validation of account type and auth method (common for all plugins)
      // will be performed into SecurityManager.runCustomServerHook
      return new AuthenticationResponse(userId, AUTH_RESULT_SUCCESS);
   }

   /**
    * Always returns <code>null</code>.
    *
    * @return  Always <code>null</code>.
    */
   @Override
   public SessionListener getSessionListener()
   {
      return null;
   }

   /**
    * Returns a set of entities that this class handles. The list of entities / parameters will
    * be configured using other ways.
    *
    * @return  A set with the value of option that was configured or an empty set if option was
    *          not configured.
    */
   @Override
   public Set<String> getAuthenticationEntities()
   {
      Set<String> ret = new HashSet<>();
      if (authOption != null)
      {
         ret.add(authOption);
      }
      return ret;
   }

   /**
    * Configures the Authenticator by setting the "option" parameter from directory.xml.
    *
    * @param   option
    *          The value of "option" entry for the auth plugin.
    */
   @Override
   public void configure(String option)
   {
      this.authOption = option;
   }
}
